/*
	Author	: Vincent Sweeney
	Email	: vince.sweeney@dxi-easycall.net
	Version : 0.4.5
	----
	Copyright (C) 2010	Vincent Sweeney
	----
*/

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <pthread.h>
#include <sys/cdefs.h>
#include <limits.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <stdarg.h>
#include <getopt.h>
#include <string.h>
#include <dirent.h>
#include <fnmatch.h>
#include <regex.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>

#if defined(__linux__)
#include <linux/version.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,13)
  #define HAVE_INOTIFY 1
#endif
#endif

#if _BSD_SOURCE || __BSD_VISIBLE || (_XOPEN_SOURCE && _XOPEN_SOURCE < 500)
  #define HAVE_DAEMON 1
#endif

#ifdef HAVE_INOTIFY
#include <sys/inotify.h>

#define EVENT_SIZE  (sizeof (struct inotify_event))
#define BUF_LEN     (1024 * (EVENT_SIZE + 16))
#endif /* HAVE_INOTIFY */

#define VERSION			"0.4.5"

/*
================
globals
================
*/

struct fifo_item_s
{
	struct fifo_item_s	*next;
	void *object;
};

struct fifo_s
{
	pthread_mutex_t mutex;
	struct fifo_item_s	*first;
	struct fifo_item_s	*last;
	int	nitems;
};

static struct fifo_s Files;
static struct fifo_s Watches;
static int nWorkers = 0;

static void scan_directory (const char *dir);

/*
================
arguments
================
*/

struct globalArgs_s
{
	struct fifo_s Directories;	/* -d option */

	char *command;		/* -c option */
	char *match;		/* --match option */
	int match_mode;		/* 0 = glob, 1 = regex */
	char *exclude;		/* --exclude option */
	int exclude_mode;	/* 0 = glob, 1 = regex */
	int ncpu;			/* -n option */
	int recursion;		/* -r option */
	int daemon;			/* --daemon option */
	int monitor;		/* -m option */
	int inotify_mask;	/* -e option */
	int only_new;
	int loop;

	off_t min_filesize;	/* --min-filesize */
	off_t max_filesize;	/* --max-filesize */

	char *logfile;		/* -l option */
	char *pidfile;		/* -p option */
	int quiet;			/* -q option */
	int verbose;		/* -v option */
} globalArgs;

static const char *optString = "d:c:n:rme:l:p:qvh";

static const struct option longOpts[] =
{
	{ "directory",		required_argument,	NULL,	'd' },
	{ "command",		required_argument,	NULL,	'c' },
	{ "nthreads",		required_argument,	NULL,	'n' },
	{ "recursion",		no_argument,		NULL,	'r' },
#ifdef HAVE_INOTIFY
	{ "monitor",		no_argument,		NULL,	'm' },
	{ "event",			required_argument,	NULL,	'e' },
	{ "only-new-files",	no_argument,		NULL,	0 },
#endif /* HAVE_INOTIFY */
	{ "loop-wait",		required_argument,	NULL,	0 },
#ifdef HAVE_DAEMON
	{ "daemon",			no_argument,		NULL,	0 },
#endif /* HAVE_DAEMON */
	{ "logfile",		required_argument,	NULL,	'l' },
	{ "pidfile",		required_argument,	NULL,	'p' },
	{ "match",			required_argument,	NULL,	0 },
	{ "match-regex",	required_argument,	NULL,	0 },
	{ "exclude",		required_argument,	NULL,	0 },
	{ "exclude-regex",	required_argument,	NULL,	0 },
	{ "min-filesize",	required_argument,	NULL,	0 },
	{ "max-filesize",	required_argument,	NULL,	0 },
	{ "quiet",			no_argument,		NULL,	'q' },
	{ "verbose",		no_argument,		NULL,	'v' },
	{ "version",		no_argument,		NULL,	0 },
	{ "help",			no_argument,		NULL,	'h' },
	{ NULL, no_argument, NULL, 0 }
};

/*
================
display_usage
================
*/

static void display_usage (int rc)
{
	printf("Usage: spawn [ options ]\n\nOptions are:\n\n");

	printf("%-22s %s", "-c | --command",
		"Command to execute. The program and any arguments should be inside single quotes.\n");
	printf("%-22s %s", "", "The string '{}' is replaced by the current filename being processed.\n");
	printf("%-22s %s", "", "The string '[]' is replaced by the current filename without its file extension.\n");
	printf("%-22s %s", "", "If neither {} or [] is used, the filename is automatically appended to the command.\n");

	printf("%-22s %s", "-d | --directory",
		"Directory to process or monitor. Defaults to current working directory.\n");

	printf("%-22s %s", "-n | --nthreads",
		"The maximum number of worker threads to spawn.\n");
	printf("%-22s %s", "", "Defaults to the logical number of processors on the system.\n");

	printf("%-22s %s", "-r | --recursion",
		"Also process files in sub-directories instead of just the specified location.\n");
	printf("%-22s %s", "", "Defaults to no recursion.\n");

#ifdef HAVE_INOTIFY
	printf("%-22s %s", "-m | --monitor",
		"Do not exit after processing the directory, instead endlessly monitor the location for new files.\n");

	printf("%-22s %s", "-e | --event",
		"Trigger on specified monitor events. Give the option more than once to use multiple events.\n");
	printf("%-22s %s", "", "Available events: access attrib close close_write close_nowrite create delete modify moved_from moved_to open\n");
	printf("%-22s %s", "", "Default: close_write. See inotify(7) for more details.\n");

	printf("%-22s %s", "   | --only-new-files",
		"When running in monitor mode, do not process files already in the directory.\n");

	printf("%-22s %s", "   | --loop-wait",
		"When not running in monitor mode, do not exit after processing the directory.\n");
	printf("%-22s %s", "", "Instead scan the directory again after the given number of seconds.\n");

#ifdef HAVE_DAEMON
	printf("%-22s %s", "   | --daemon",
		"Detach from the console and run the program in daemon mode.\n");
	printf("%-22s %s", "", "Only valid if the --monitor or --loop-wait options are enabled.\n");
#endif /* HAVE_DAEMON */

#else
	printf("%-22s %s", "   | --loop-wait",
		"Do not exit after processing the directory.\n");
	printf("%-22s %s", "", "Instead scan the directory again after the given number of seconds.\n");

#ifdef HAVE_DAEMON
	printf("%-22s %s", "   | --daemon",
		"Detach from the console and run the program in daemon mode.\n");
	printf("%-22s %s", "", "Only valid if the --loop-wait option is enabled.\n");
#endif /* HAVE_DAEMON */

#endif /* HAVE_INOTIFY */

	printf("%-22s %s", "-l | --logfile",
		"Enable logging of all console output to the given file. Output is appended to any existing file.\n");

	printf("%-22s %s", "-p | --pidfile",
		"Write main process id to file.\n");

	printf("%-22s %s", "   | --match",
		"Only process files matching the specified glob(3) pattern.\n");

	printf("%-22s %s", "   | --match-regex",
		"Only process files matching the specified regex(3) pattern.\n");

	printf("%-22s %s", "   | --exclude",
		"Ignore files matching the specified glob(3) pattern.\n");

	printf("%-22s %s", "   | --exclude-regex",
		"Ignore files matching the specified regex(3) pattern.\n");
	printf("%-22s %s", "", "Files are excluded before being matched. Patterns are matched against the full path + filename.\n");

	printf("%-22s %s", "   | --min-filesize",
		"Only process files at least the specified filesize, in bytes.\n");

	printf("%-22s %s", "   | --max-filesize",
		"Only process files at most the specified filesize, in bytes.\n");

	printf("%-22s %s", "-q | --quiet",
		"Disable all console output. Output to the logfile is unaffected.\n");

	printf("%-22s %s", "-v | --verbose",
		"Print more detailed output. This option can be specified more than once.\n");

	printf("%-22s %s", "   | --version",
		"Display program version details.\n");

	printf("%-22s %s", "-h | --help",
		"Display this message.\n");

	printf("\n");
	printf("Example - recursively backup all files in the current directory: spawn -r -c 'cp -f {} {}~'\n");
	printf("Example - parallel convert all .wav files to .mp3: spawn -c 'lame {} [].mp3' --match \"*.wav\"\n");
	printf("\n");

	exit(rc);
}

/*
================
lprintf
================
*/

static void lprintf (const char *msg, ...)
{
	va_list argptr;

	if (!globalArgs.quiet)
	{
		va_start(argptr, msg);
		vfprintf(stdout, msg, argptr);
		va_end(argptr);
	}

	if (globalArgs.logfile)
	{
		va_start(argptr, msg);
		FILE *f = fopen(globalArgs.logfile, "a");
		if (!f)
		{
			perror("Unable to write to logfile.");
			exit(EXIT_FAILURE);
		}
		vfprintf(f, msg, argptr);
		fclose(f);
		va_end(argptr);
	}
}

/*
================
display_version
================
*/

static void display_version (int quit)
{
	if (quit)
	{
		printf("spawn v%s\n", VERSION);
		printf("Author: Vincent Sweeney <vince.sweeney@dxi-easycall.net>\n");
		exit(EXIT_SUCCESS);
	}
	else
	{
		lprintf("spawn v%s\n", VERSION);
		lprintf("Author: Vincent Sweeney <vince.sweeney@dxi-easycall.net>\n");
	}
}

/*
================
write_pid_file
================
*/

static void write_pid_file (const char *fname, pid_t pid)
{
	if (pid)
	{
		FILE *f = fopen(fname, "w");
		if (!f)
		{
			perror("Unable to write pid file.");
			exit(EXIT_FAILURE);
		}
		fprintf(f, "%u\n", pid);
		fclose(f);
	}
	else
		remove(fname);
}

/*
================
is_empty_string
================
*/

static int is_empty_string (const char *str)
{
	char ch;

	if (str != NULL)
		while ( (ch = *str++) )
			if (!isspace(ch)) return 0;

	return 1;
}

/*
================
build_command
================
*/

static char *build_command (char *cmd, const char *src, const char *fname)
{
	char *dst = cmd;
	int found = 0;
	char ch;

	while ((ch = *src++))
	{
		switch(ch)
		{
		case '\\':
			*dst++ = ch;
			if (!(*dst++ == *src++))
				break;
			continue;
		case '{':
			if (*src == '}')
			{
				found = 1;
				strcpy(dst, fname);
				dst += strlen(fname);
				src++;
			}
			else
				*dst++ = ch;
			break;
		case '[':
			if (*src == ']')
			{
#if defined (__linux__)
				char *delim, *str = strdupa(fname);
#else
				char *delim, *str = alloca(strlen(fname) + 1);
				strcpy(str, fname);
#endif
				delim = strrchr(str, '.');
				if (delim) *delim = '\0';

				found = 1;
				strcpy(dst, str);
				dst += strlen(str);
				src++;
			}
			else
				*dst++ = ch;
			break;
		default:
			*dst++ = ch;
			break;
		}
	}

	if (!found)
	{
		*dst++ = ' ';
		strcpy(dst, fname);
	}
	else
		*dst++ = '\0';

	return cmd;
}

/*
================
rgmatch
================
*/

static int rgmatch (const char *match, const char *str)
{
	regex_t reg;
	int rc;

	rc = regcomp(&reg, globalArgs.match, REG_NOSUB);
	if (rc < 0)
	{
		char errstr[256];

		regerror(rc, &reg, errstr, sizeof(errstr));
		lprintf("Regex error: %s\n", errstr);
		exit(EXIT_FAILURE);
	}

	rc = regexec(&reg, str, 0, NULL, 0);
	regfree(&reg);

	if (rc == REG_NOMATCH)
		return 1;

	return 0;
}

/*
================
match_fullpath
================
*/

static int match_fullpath (char *path)
{
	/* Match filename */
	if (!is_empty_string(globalArgs.exclude))
	{
		if (globalArgs.exclude_mode == 0)
		{
			/* Glob */
			if (fnmatch(globalArgs.exclude, path, 0) == 0)
				return 1;
		}
		else
		{
			/* Regex */
			if (rgmatch(globalArgs.exclude, path) == 0)
				return 1;
		}
	}

	if (!is_empty_string(globalArgs.match))
	{
		if (globalArgs.match_mode == 0)
		{
			/* Glob */
			if (fnmatch(globalArgs.match, path, 0) != 0)
				return 1;
		}
		else
		{
			/* Regex */
			if (rgmatch(globalArgs.match, path) != 0)
				return 1;
		}
	}

	return 0;
}

/*
================
fifo_push
================
*/

static void fifo_push (struct fifo_s *List, void *obj, size_t len)
{
	struct fifo_item_s *item;

	item = malloc(sizeof(*item) + len);
	if (!item)
	{
		perror("malloc()");
		exit(EXIT_FAILURE);
	}
	item->next = NULL;
	item->object = (char *) item + sizeof(*item);
	memcpy(item->object, obj, len);

	pthread_mutex_lock(&List->mutex);
	if (!List->nitems)
		List->first = item;
	else
		List->last->next = item;
	List->last = item;
	List->nitems++;
	pthread_mutex_unlock(&List->mutex);
}

/*
================
fifo_pop
================
*/

static struct fifo_item_s *fifo_pop (struct fifo_s *List)
{
	struct fifo_item_s *item = NULL;

	pthread_mutex_lock(&List->mutex);
	if (List->nitems && (item = List->first))
	{
		if (--List->nitems == 0)
			List->first = List->last = NULL;
		else
			List->first = item->next;
	}
	pthread_mutex_unlock(&List->mutex);

	return item;
}

/*
================
init_fifo
================
*/

static void init_fifo (struct fifo_s *List)
{
	pthread_mutex_init(&List->mutex, NULL);
	List->first = NULL;
	List->last  = NULL;
	List->nitems = 0;
}

/*
================
file_check
================
*/

int file_check (const char *path, const char *fname, int scandir)
{
	struct stat st;

	if (stat(path, &st) < 0) return 1;
	if (scandir && S_ISDIR(st.st_mode))
	{
		if (strcmp(fname, ".") == 0 || strcmp(fname, "..") == 0)
			return 1;

		if (globalArgs.recursion)
			scan_directory(path);

		return 1;
	}
	if (!S_ISREG(st.st_mode)) return 1;

	/* Check filesize */
	if (globalArgs.max_filesize > 0 && st.st_size > globalArgs.max_filesize)
		return 1;
	if (globalArgs.min_filesize > 0 && st.st_size < globalArgs.min_filesize)
		return 1;

	return 0;
}

/*
================
scan_directory
================
*/

static void scan_directory (const char *dir)
{
	char path[PATH_MAX];

	struct dirent **namelist;
	char *fname;
	int i, n;

	if (globalArgs.verbose >= 3) lprintf("Scanning directory: %s\n", dir);

	/* Scan Directory */
	n = scandir(dir, &namelist, 0, alphasort);
	if (n <= 0) return;

	for (i = 0; i < n; i++)
	{
		fname = namelist[i]->d_name;
		snprintf(path, sizeof(path), "%s/%s", dir, fname);

		/* Check the filename */
		if (file_check(path, fname, 1) != 0) continue;

		/* Match the filename */
		if (match_fullpath(path) != 0) continue;

		/* Store file to process */
		fifo_push(&Files, path, strlen(path) + 1);
	}

	for (i = 0; i < n; i++) free(namelist[i]);
	free(namelist);
}

#ifdef HAVE_INOTIFY

struct dirWatches_s
{
	char *path;
	int wd;
};

/*
================
event_string
================
*/

static char *event_string (char *buf, int mask)
{
	char *str = buf;

	*str = '\0';
	if (mask & IN_ACCESS)			str += sprintf(str, " access");
	if (mask & IN_ATTRIB)			str += sprintf(str, " attrib");
	if (mask & IN_CLOSE_WRITE)		str += sprintf(str, " close_write");
	if (mask & IN_CLOSE_NOWRITE)	str += sprintf(str, " close_nowrite");
	if (mask & IN_CREATE)			str += sprintf(str, " create");
	if (mask & IN_DELETE)			str += sprintf(str, " delete");
	if (mask & IN_MODIFY)			str += sprintf(str, " access");
	if (mask & IN_MOVED_FROM)		str += sprintf(str, " moved_from");
	if (mask & IN_MOVED_TO)			str += sprintf(str, " moved_to");
	if (mask & IN_OPEN)				str += sprintf(str, " open");

	return buf + 1;
}

/*
================
watch_directory
================
*/

static void watch_directory (int ifd, char *dir)
{
	struct dirWatches_s watch;
	char path[PATH_MAX];

	struct dirent **namelist;
	struct stat st;
	char *fname;
	int i, n, wd;

	wd = inotify_add_watch(ifd, dir, globalArgs.inotify_mask);
	if (wd < 0)
	{
		lprintf("Error adding watch to directory: %s\n", dir);
		return;
	}
	else if (globalArgs.verbose >= 2)
		lprintf("Monitoring directory [%d]: %s\n", wd, dir);

	watch.wd = wd;
	watch.path = strdup(dir);
	fifo_push(&Watches, &watch, sizeof(watch));

	if (!globalArgs.recursion) return;

	/* Watch Sub-Directories */
	n = scandir(dir, &namelist, 0, alphasort);
	if (n <= 0) return;

	for (i = 0; i < n; i++)
	{
		fname = namelist[i]->d_name;
		snprintf(path, sizeof(path), "%s/%s", dir, fname);
		if (stat(path, &st) < 0) continue;
		if (!S_ISDIR(st.st_mode)) continue;
		if (strcmp(fname, ".") == 0 || strcmp(fname, "..") == 0) continue;

		watch_directory(ifd, path);
	}

	for (i = 0; i < n; i++) free(namelist[i]);
	free(namelist);
}

/*
================
inotify_thread
================
*/

static void *inotify_thread (void *arg)
{
	struct fifo_item_s *item;

	char buffer[BUF_LEN];
	char path[PATH_MAX], evbuf[256];
	int i, length;

	int ifd = inotify_init();
	if (ifd < 0)
	{
		globalArgs.monitor = 0;
		perror("inotify_init");
		pthread_exit(NULL);
	}

	for (item = globalArgs.Directories.first; item; item = item->next)
		watch_directory(ifd, item->object);
	fflush(NULL);

	while (1)
	{
		struct inotify_event *event;

		/* Wait for file close notifications */
		length = read(ifd, buffer, BUF_LEN);
		if (length < 0)
		{
			if (errno == EINTR) continue;
			globalArgs.monitor = 0;
			perror("inotify_read");
			break;
		}

		for (i = 0; i < length; i += EVENT_SIZE + event->len)
		{
			struct dirWatches_s *watch;

			event = (struct inotify_event *) &buffer[i];

			if (!event->len) continue;
			if (!(event->mask & globalArgs.inotify_mask) || (event->mask & IN_ISDIR)) continue;

			for (item = Watches.first; item; item = item->next)
			{
				watch = item->object;
				if (watch->wd == event->wd)
				{
					sprintf(path, "%s/%s", watch->path, event->name);

					/* Check the filename */
					if (file_check(path, event->name, 0) == 0 && match_fullpath(path) == 0)
					{
						if (globalArgs.verbose >= 2)
							lprintf("Event triggered [%s] %s\n", event_string(evbuf, event->mask), path);

						/* Store file to process */
						fifo_push(&Files, path, strlen(path) + 1);
					}
					break;
				}
			}
		}
		fflush(NULL);
	}

	close(ifd);
	pthread_exit(NULL);
}
#endif /* HAVE_INOTIFY */

/*
================
main
================
*/

int main(int argc, char *argv[])
{
	int longIndex, opt, pid, status;
	struct fifo_item_s *item;
	char *path;

	/* Defaults */
	init_fifo(&Files);
	init_fifo(&Watches);
	init_fifo(&globalArgs.Directories);
	globalArgs.command = NULL;
	globalArgs.pidfile = NULL;
	globalArgs.logfile = NULL;
	globalArgs.match = NULL;
	globalArgs.match_mode = 0;
	globalArgs.exclude = NULL;
	globalArgs.exclude_mode = 0;
	globalArgs.min_filesize = 0;
	globalArgs.max_filesize = 0;
	globalArgs.ncpu = 0;
	globalArgs.recursion = 0;
	globalArgs.monitor = 0;
	globalArgs.inotify_mask = 0;
	globalArgs.loop = 0;
	globalArgs.daemon = 0;
	globalArgs.quiet = 0;
	globalArgs.verbose = 0;
	globalArgs.only_new = 0;

	/* Parse the command line */
	while (1)
	{
		opt = getopt_long(argc, argv, optString, longOpts, &longIndex);
		if (opt == -1) break;

		switch (opt)
		{
		case 'd':
			{
				char buf[PATH_MAX];
				path = realpath(optarg, buf);
				if (path == NULL)
				{
					perror("Unable to expand directory location.");
					exit(EXIT_FAILURE);
				}
				fifo_push(&globalArgs.Directories, path, strlen(path) + 1);
			}
			break;
		case 'c':
			globalArgs.command = optarg;
			break;
		case 'p':
			globalArgs.pidfile = optarg;
			break;
		case 'l':
			globalArgs.logfile = optarg;
			break;
		case 'n':
			globalArgs.ncpu = atoi(optarg);
			break;
		case 'r':
			globalArgs.recursion = 1;
			break;
		case 'q':
			globalArgs.quiet++;
			break;
		case 'v':
			globalArgs.verbose++;
			break;
		case 'h':
			display_usage(EXIT_SUCCESS);
			break;
		case '?':
			display_usage(EXIT_FAILURE);
			break;
#ifdef HAVE_INOTIFY
		case 'm':
			globalArgs.monitor = 1;
			break;
		case 'e':
			if (strcmp(optarg, "access") == 0)
				globalArgs.inotify_mask |= IN_ACCESS;
			else if (strcmp(optarg, "attrib") == 0)
				globalArgs.inotify_mask |= IN_ATTRIB;
			else if (strcmp(optarg, "close") == 0)
				globalArgs.inotify_mask |= IN_CLOSE;
			else if (strcmp(optarg, "close_write") == 0)
				globalArgs.inotify_mask |= IN_CLOSE_WRITE;
			else if (strcmp(optarg, "close_nowrite") == 0)
				globalArgs.inotify_mask |= IN_CLOSE_NOWRITE;
			else if (strcmp(optarg, "create") == 0)
				globalArgs.inotify_mask |= IN_CREATE;
			else if (strcmp(optarg, "delete") == 0)
				globalArgs.inotify_mask |= IN_DELETE;
			else if (strcmp(optarg, "modify") == 0)
				globalArgs.inotify_mask |= IN_MODIFY;
			else if (strcmp(optarg, "moved_from") == 0)
				globalArgs.inotify_mask |= IN_MOVED_FROM;
			else if (strcmp(optarg, "moved_to") == 0)
				globalArgs.inotify_mask |= IN_MOVED_TO;
			else if (strcmp(optarg, "open") == 0)
				globalArgs.inotify_mask |= IN_OPEN;
			else
			{
				lprintf("Unknown event: %s\n", optarg);
				display_usage(EXIT_FAILURE);
			}
			break;
#endif /* HAVE_INOTIFY */
		case 0:
			/* long option without a short arg */
			if (strcmp("version", longOpts[longIndex].name) == 0)
			{
				display_version(1);
			}
			else if (strcmp("match", longOpts[longIndex].name) == 0)
			{
				globalArgs.match = optarg;
				globalArgs.match_mode = 0;
			}
			else if (strcmp("match-regex", longOpts[longIndex].name) == 0)
			{
				globalArgs.match = optarg;
				globalArgs.match_mode = 1;
			}
			else if (strcmp("exclude", longOpts[longIndex].name) == 0)
			{
				globalArgs.exclude = optarg;
				globalArgs.exclude_mode = 0;
			}
			else if (strcmp("exclude-regex", longOpts[longIndex].name) == 0)
			{
				globalArgs.exclude = optarg;
				globalArgs.exclude_mode = 1;
			}
			else if (strcmp("loop-wait", longOpts[longIndex].name) == 0)
			{
				globalArgs.loop = atoi(optarg);
			}
			else if (strcmp("min-filesize", longOpts[longIndex].name) == 0)
			{
				globalArgs.min_filesize = atoll(optarg);
			}
			else if (strcmp("max-filesize", longOpts[longIndex].name) == 0)
			{
				globalArgs.max_filesize = atoll(optarg);
			}
#ifdef HAVE_INOTIFY
			else if (strcmp("only-new-files", longOpts[longIndex].name) == 0)
			{
				globalArgs.only_new = 1;
			}
#endif /* HAVE_INOTIFY */
#ifdef HAVE_DAEMON
			else if (strcmp("daemon", longOpts[longIndex].name) == 0)
			{
				globalArgs.daemon = 1;
			}
			break;
#endif /* HAVE_DAEMON */
		default:
			break;
		}
	}

	/* Sanity check the arguments */
	if (is_empty_string(globalArgs.command))
	{
		fprintf(stderr, "No command given.\n");
		display_usage(EXIT_FAILURE);
	}

	if (globalArgs.Directories.nitems == 0)
	{
		path = realpath(".", NULL);
		fifo_push(&globalArgs.Directories, path, strlen(path) + 1);
		free(path);
	}

	if (globalArgs.ncpu < 1)
	{
		globalArgs.ncpu = sysconf(_SC_NPROCESSORS_ONLN);
		if (globalArgs.ncpu < 1)
		{
			perror("Unable to determine number of processors.");
			exit(EXIT_FAILURE);
		}
	}

#ifdef HAVE_DAEMON
	if (globalArgs.daemon)
	{
		if (!globalArgs.monitor && !globalArgs.loop)
		{
			fprintf(stderr, "daemon mode requested but --monitor or --loop-wait options not set.\n");
			exit(EXIT_FAILURE);
		}
		globalArgs.quiet = 1;
		daemon(1, 0);
	}
#endif /* HAVE_DAEMON */

	/* Print settings */
	if (globalArgs.verbose)
	{
		display_version(0); printf("\n");
		lprintf("Verbose Level: %d\n", globalArgs.verbose);
		for (item = globalArgs.Directories.first; item; item = item->next)
			lprintf("Processing Directory: %s\n", item->object);
		lprintf("Recursion: %s\n", globalArgs.recursion ? "enabled" : "disabled");
		lprintf("Spawning upto %d worker threads.\n", globalArgs.ncpu);
		if (is_empty_string(globalArgs.match))
			lprintf("Matching all files.\n");
		else
			lprintf("Matching files against %s pattern: %s\n", globalArgs.match_mode ? "regex" : "glob", globalArgs.match);
		fflush(NULL);
	}

	/* Build file list */
	if (!globalArgs.monitor || !globalArgs.only_new)
	{
		for (item = globalArgs.Directories.first; item; item = item->next)
			scan_directory(item->object);
		if (globalArgs.verbose)
		{
			lprintf("Found %d files to process.\n", Files.nitems);
			fflush(NULL);
		}
	}
	else
	{
		if (globalArgs.verbose)
		{
			lprintf("Skipping initial directory scan.\n");
			fflush(NULL);
		}
	}
#ifdef HAVE_INOTIFY
	pthread_t ithread;

	if (globalArgs.inotify_mask == 0) globalArgs.inotify_mask = IN_CLOSE_WRITE;

	if (globalArgs.monitor)
	{
		char evbuf[256];
		int rc;

		rc = pthread_create(&ithread, NULL, &inotify_thread, NULL);
		if (rc != 0)
		{
			perror("Unable to start monitor thread.");
			exit(EXIT_FAILURE);
		}

		if (globalArgs.verbose)
		{
			lprintf("Monitor thread started.\n");
			lprintf("Triggering on events: %s\n", event_string(evbuf, globalArgs.inotify_mask));
			fflush(NULL);
		}
	}
#endif /* HAVE_INOTIFY */

	if (globalArgs.pidfile) write_pid_file(globalArgs.pidfile, getpid());

	/* Main process loop */
	while (globalArgs.monitor || globalArgs.loop > 0 || Files.nitems)
	{
		struct fifo_item_s *item;

		/* Acknowledge any zombies */
		while (nWorkers)
		{
			pid = waitpid(-1, &status, WNOHANG);
			if (pid <= 0) break;
			nWorkers--;
		}

		/* Monitoring and no files left in fifo queue */
		if (Files.nitems == 0)
		{
			if (globalArgs.loop > 0 && !globalArgs.monitor)
			{
				sleep(globalArgs.loop);
				for (item = globalArgs.Directories.first; item; item = item->next)
					scan_directory(item->object);
			}
			else
				sleep(1);
			continue;
		}

		/* If we are totally busy, wait for a thread to finish */
		if (nWorkers == globalArgs.ncpu)
		{
			pid = waitpid(-1, &status, 0);
			nWorkers--;
		}

		item = fifo_pop(&Files);
		if (!item) continue;

		/* fork a new worker process */
		pid = fork();
		if (pid < 0)
		{
			perror("fork() error");
			exit(EXIT_FAILURE);
		}
		else if (pid > 0)
		{
			/* Parent process */
			nWorkers++;
			free(item);
		}
		else
		{
			/* Child process */
			char cmd[PATH_MAX];
			build_command(cmd, globalArgs.command, item->object);

			if (globalArgs.verbose >= 2) lprintf("Processing Target [%d] %s\n", Files.nitems + 1, item->object);
			if (globalArgs.verbose >= 3) lprintf("Running: %s\n", cmd);
			execl("/bin/sh", "sh", "-c", cmd, (char *) NULL);
			exit(EXIT_SUCCESS);
		}
	}

	/* Wait for all workers to finish */
	while (nWorkers)
	{
		pid = waitpid(-1, &status, 0);
		nWorkers--;
	}

	if (globalArgs.verbose) lprintf("Processing Complete.\n");
	if (globalArgs.pidfile) write_pid_file(globalArgs.pidfile, 0);

	exit(EXIT_SUCCESS);
}
/*
 * vim:ts=4:sw=4
 */
